from pyAudioKits.analyse.timeDomainAnalyse import energyCal, overzeroCal
import numpy as np
import matplotlib.pyplot as plt

class VAD:

    def __init__(self,input,energyThresLow,energyThresHigh,zerocrossingThres,frameDuration = 0.03,overlapRate=0.5):
        """Speech endpoint detection based on double threshold. 

        input: An Audio object. 
        energyThresLow: A lower threshold of energy for distinguish between silence and voice.
        energyThresHigh: A higher threshold of energy for distinguish between unvoiced and voiced.
        zerocrossingThres: Zero crossing rate threshold.
        frameDuration: A float object for the duration of each frame (seconds) or a int object for the length of each fram (sample points). 
        overlapRate: A float object in [0,1) for the overlapping rate of the frame.

        return: A VAD object.  
        """
        input = input.framing(frameDuration, overlapRate)
        flatten=input.samples
        energys=energyCal(flatten)
        overzeros=overzeroCal(flatten,input.sr)

        voice_begins, voice_ends = self.__distinguish(0, len(energys), energys, overzeros, True, energyThresLow, zerocrossingThres)

        labels = np.zeros_like(energys)
        for i in range(len(voice_begins)):
            labels[voice_begins[i]:voice_ends[i]] = 1

        for b, e in zip(voice_begins, voice_ends):
            voice_begins2, voice_ends2 = self.__distinguish(b, e, energys, overzeros, False, energyThresHigh, zerocrossingThres)
            for i in range(len(voice_begins2)):
                labels[voice_begins2[i]:voice_ends2[i]] = 2

        self.label=labels
        self.input=input
    
    def __distinguish(self, b, e, energys, overzeros, overzeroOver, energyThres, overzeroThres):
        clip1=energyThres
        voice_begins = []
        voice_ends = []
        if energys[b] >= clip1:
            voice_begins.append(b)
        for i in range(b+1,e):
            if (energys[i] >= clip1) and (energys[i-1] < clip1):
                voice_begins.append(i)
            elif (energys[i] <= clip1) and (energys[i-1] > clip1):
                voice_ends.append(i)
        if len(voice_begins) - 1 == len(voice_ends):
            voice_ends.append(e)
        assert len(voice_begins) == len(voice_ends)
        
        clip2=overzeroThres
        voice_begins2 = []
        voice_ends2 = []
        for voice_begin in voice_begins[::-1]:
            appended = False
            for i in range(voice_begin, b-1, -1):
                if (overzeroOver and (overzeros[i] < clip2)) or ((not overzeroOver) and (overzeros[i] >= clip2)):
                    voice_begins2.append(i+1)
                    appended = True
                    break
            if not appended:
                voice_begins2.append(b)
        voice_begins2 = voice_begins2[::-1]
        for voice_end in voice_ends:
            appended = False
            for i in range(voice_end, e):
                if (overzeroOver and (overzeros[i] < clip2)) or ((not overzeroOver) and (overzeros[i] >= clip2)):
                    voice_ends2.append(i)
                    appended = True
                    break
            if not appended:
                voice_ends2.append(e)
        assert len(voice_begins2) == len(voice_ends2)

        return voice_begins2, voice_ends2


    def plot(self,imgPath=None):
        """Visualize the result of VAD. Save the figure if imgPath is given, otherwise display directly. The blue parts are silence and the magenta parts are voice. 

        imgPath: The direction to save the figure. 
        """
        label=self.label
        audio=self.input.retrieve()
        step=self.input.step
        label2=np.array([label[i] for i in range(0,label.shape[0]) for n in range(0,step)])
        x=np.arange(0,label2.shape[0]/audio.sr,1/audio.sr)
        y=audio.samples[:label2.shape[0]]
        plt.xlabel("t/s")
        plt.ylabel("amplitude")
        plt.plot(x,np.where(label2==0,y,None))
        plt.plot(x,np.where(label2==1,y,None),color="m")
        plt.plot(x,np.where(label2==2,y,None),color="orange")
        if imgPath!=None:
            plt.savefig(imgPath,dpi=500, bbox_inches = 'tight')
        else:
            plt.show()
        plt.clf()

    def slices(self):
        """Return the slices of voices in audio. 

        return: A list for Audio objects. 
        """
        label=self.label
        audio=self.input.retrieve()
        step=self.input.step
        slices=[]
        label2=[]
        start=0
        for i in range(0,len(label)-1):
            if (label[i+1]==0 and label[i]!=0) or (label[i+1]!=0 and label[i]==0):
                slices.append(audio[start*step/audio.sr:(i+1)*step/audio.sr])
                if label[i]==0:
                    label2.append(0)
                else:
                    label2.append(1)
                start=i+1
        return [s for s, l in zip(slices, label2) if l==1]


from scipy.linalg import toeplitz
from pyAudioKits.audio.audio import AudioFrames,Audio,srJudge
import numpy as np
from pyAudioKits.analyse.modelBasedAnalyse import LPC,LPC1time,nextpow2
import scipy.signal

def specSubstract(input, noise, beta=0.002, frameDuration = 0.03, overlapRate = 0.5, window = None):
    """Using spectral subtraction to reduce noise. 

    input: An Audio object of signal + noise. 
    noise: An Audio object of estimate noise. 
    beta: The beta parameter. 
    frameDuration: A float object for the duration of each frame (seconds) or a int object for the length of each fram (sample points). 
    overlapRate: A float object in [0,1) for the overlapping rate of the frame. 
    window: 
            If string, it's the name of the window function (e.g., "hann")
            If tuple, it's the name of the window function and any parameters (e.g., ("kaiser", 4.0))
            If numeric, it is treated as the beta parameter of the "kaiser" window, as in scipy.signal.get_window.
            If callable, it's a function that accepts one integer argument (the window length)
            If list-like, it's a pre-computed window of the correct length Nx

    return: An Audio object of filtered signal. 
    """
    srJudge(input, noise)
    input = input.framing(frameDuration, overlapRate, window)
    noise = noise.framing(frameDuration, overlapRate, window)
    x=input.samples.T
    len_ = input.samples.shape[1]
    nFFT = 2 * 2 ** (nextpow2(len_))
    Expnt=2
    noise_mu = np.abs(np.fft.fft(noise.samples,axis=-1,n=nFFT))
    if len(noise_mu.shape) > 1:
        noise_mu=np.mean(noise_mu,axis=0)
    noise_mu = noise_mu[:,None]

    img = 1j
    spec = np.fft.fft(x, n=nFFT,axis=0)
    sig=np.abs(spec)
    theta=np.angle(spec)
    SNRseg = 10 * np.log10(np.linalg.norm(sig, 2,axis=0) ** 2 / np.linalg.norm(noise_mu, 2,axis=0) ** 2)
    def berouti(SNR):
        SNR=np.where((SNR>=-5)&(SNR<=20),4 - SNR * 3 / 20,SNR)
        SNR=np.where(SNR<-5,5,SNR)
        SNR=np.where(SNR>20,1,SNR)
        return SNR
    alpha = berouti(SNRseg)
    sub_speech = sig ** Expnt - alpha * noise_mu ** Expnt
    diffw = sub_speech - beta * noise_mu ** Expnt
    sub_speech=np.where(diffw<0,beta*noise_mu**Expnt,sub_speech)
    x_phase = (sub_speech ** (1 / Expnt)) * (np.cos(theta) + img * np.sin(theta))
    xi = np.fft.ifft(x_phase,axis=0).real
    xi = xi[:input.samples.shape[1]]
    return AudioFrames(xi.T,input.sr,input.step,input.window).retrieve()

def wienerFilter(observed_signal,desired_signal,h_length=200, frameDuration = 0.03, overlapRate = 0.5, window = None):
    """Using Wiener filtering to reduce noise. 

    observed_signal: An Audio object of signal + noise. 
    desired_signal: An Audio object or estimated signal. 
    h_length: Orders. 
    frameDuration: A float object for the duration of each frame (seconds) or a int object for the length of each fram (sample points). 
    overlapRate: A float object in [0,1) for the overlapping rate of the frame. 
    window: 
            If string, it's the name of the window function (e.g., "hann")
            If tuple, it's the name of the window function and any parameters (e.g., ("kaiser", 4.0))
            If numeric, it is treated as the beta parameter of the "kaiser" window, as in scipy.signal.get_window.
            If callable, it's a function that accepts one integer argument (the window length)
            If list-like, it's a pre-computed window of the correct length Nx

    return: An Audio object of filtered signal. 
    """
    srJudge(observed_signal, desired_signal)
    observed_signal = observed_signal.framing(frameDuration, overlapRate, window)
    desired_signal = desired_signal.framing(frameDuration, overlapRate, window)
    row_number = h_length
    col_number = row_number
    M = col_number
    results = []
    for i in range(observed_signal.shape[0]):
        noise_=desired_signal.samples[i]
        observed_signal_=observed_signal.samples[i]
        Rx_c_full = scipy.signal.correlate(observed_signal_,observed_signal_,"full")
        k = np.argmax(Rx_c_full)
        Rx_c = Rx_c_full[k:k+M]
        Rdx_c_full = scipy.signal.correlate(noise_,observed_signal_,"full")
        Rdx_c = Rdx_c_full[k:k+M]
        A = toeplitz(Rx_c,Rx_c)
        b = np.mat(Rdx_c[:,None])
        h = np.mat(A).I*b
        h=np.squeeze(np.array(h))
        #results.append(Audio(observed_signal_ - scipy.signal.convolve(h, observed_signal_,"full")[:len(observed_signal_)],observed_signal.sr))
        results.append(scipy.signal.convolve(h, observed_signal_,"full")[:len(observed_signal_)])
    return AudioFrames(np.array(results), observed_signal.sr, observed_signal.step, observed_signal.window).retrieve()


def kalmanFilter(input,noise,numIter=7,p=20, frameDuration = 0.05):
    """Using Kalman filtering to reduce noise. 

    input: An Audio object of signal + noise.
    noise: An Audio object of estimate noise.   
    numIter: Iterating times. 
    p: Orders. 
    frameDuration: A float object for the duration of each frame (seconds) or a int object for the length of each fram (sample points). 

    return: An Audio object of filtered signal. 
    """    
    srJudge(input, noise)
    f = input.framing(frameDuration,0)
    framedSignal=f.samples
    H=np.zeros([1,p])
    H[0,-1]=1
    H=np.mat(H)
    R = np.var(noise.samples)
    Q,lpcs=LPC(f,p)
    errCov = np.mat(R * np.eye(p))
    output = np.zeros(input.samples.shape[0])
    output[0:p] = input.samples[0 : p]
    estOutput = np.mat(input.samples[:p][:,None])
    part1=np.eye(p-1)
    part0=np.zeros([p-1,1])
    part=np.concatenate([part0,part1],axis=1)
    w=framedSignal.shape[1]
    for k in range(framedSignal.shape[0]):
        oldOutput = estOutput
        if k == 0:
            iiStart = p
        else:
            iiStart = 0
        for iter in range(numIter):
            A=np.mat(np.concatenate([part,np.flip(-lpcs[k])[None]],axis=0))
            for ii in range(iiStart,w):
                aheadEstOutput = A * estOutput
                aheadErrCov  = A * errCov * A.T + H.T * Q[k] * H
                K = (aheadErrCov * H.T) / (H * aheadErrCov * H.T + R)
                estOutput = aheadEstOutput + K * (framedSignal[k, ii] - H * aheadEstOutput)
                index = ii - iiStart + p  + k * w
                output[index - p  : index] = np.squeeze(estOutput)
                errCov  = (np.mat(np.eye(p)) - K * H) * aheadErrCov
            if iter < numIter-1:
                estOutput = oldOutput
            Q[k],lpcs[k] = LPC1time(output[k * w  : (k+1) * w], p)
    f1=Audio(output,input.sr)
    return f1



from .axis import *
import matplotlib.pyplot as plt
import pandas as pd
import seaborn as sns
import librosa

def tResult(values, input, unit):
    if type(values) == np.ndarray:
        return frame_axis_ds(values, input.sr, input.step, unit)
    else:
        return values

def fResult(values, input, unit):
    if unit == "complex":
        if len(values.shape) == 2:
            return freq_frame_axis_ds(values, input.sr, input.step, unit)
        else:
            return freq_axis_ds(values, input.sr, unit)
    else:
        if len(values.shape) == 2:
            return freq_frame_axis_ds_real(values, input.sr, input.step, unit)
        else:
            return freq_axis_ds_real(values, input.sr, unit)

class one_axis_ds:
    def __init__(self, samples, sr, axis, unit):
        self.samples = samples
        self.sr = sr
        self.axis = axis
        self.shape = samples.shape
        self.unit = unit
    
    def __getitem__(self, index):
        if type(index) == tuple:
            raise Exception("Only support 1-D indices. ")
        else:
            return self.samples[self.axis[index]]
    
    def __len__(self):
        return self.samples.shape[0]
    
    def plot(self,start=0,end=None,ylim=None,ax=None,imgPath=None,xlabel="t/s",plot_transform=lambda x:x,slicing=lambda a,b,c,d,e:e.slicing(a,b,c,d),unit=None):
        indice = slicing(start,end,None,xlabel,self.axis)
        y = self.samples[indice]
        x = self.axis.getScale(xlabel)[indice]
        if ax==None:
            plt.xlabel(xlabel)
            if unit is None:
                plt.ylabel(self.unit)
            else:
                plt.ylabel(unit)
            if not ylim is None:
                plt.ylim(ylim[0],ylim[1])
            plt.plot(x, plot_transform(y))
            if imgPath==None:
                plt.show()
            else:
                plt.savefig(imgPath,dpi=500, bbox_inches = 'tight')
            plt.clf()
        else:
            ax.plot(x, y)
    
class time_axis_ds(one_axis_ds):
    def __init__(self, samples, sr, unit):
        super().__init__(samples, sr, time_axis(len(samples), sr), unit)
    
    def plot(self, start=0, end=None, ylim=None, ax=None, imgPath=None, xlabel="t/s"):
        """To draw the audio waveform on the sub graph. 
        If no subgraph is passed in, it will be displayed directly. 
        If imgpath is passed in, the graph will be saved. 

        start: 
            If xlabel is "t/s" or "t/ms", then it will be the starting time stamp. Default = 0. 
            If xlabel is "n", then it will be the starting sample count. Default = 0.
        end: 
            If xlabel is "t/s" or "t/ms", then it will be the ending time stamp. Default = The last time stamp of the audio. 
            If xlabel is "n", then it will be the ending sample count. Default = The total count of samples. 
        ylim: A tuple (y_start,y_end) for display range on y-axis. The default (None) is adaptive. 
        ax: A matplotlib.pyplot subplot to draw on. 
        imgPath: The path to save the graph. 
        xlabel: "t/s", "t/ms" or "n". 
        """
        return super().plot(start, end, ylim, ax, imgPath, xlabel)
        

class frame_axis_ds(one_axis_ds):
    def __init__(self, samples, sr, step, unit):
        super().__init__(samples, sr, frame_axis(len(samples), sr, step), unit)
    
    def plot(self, start=0, end=None, ylim=None, ax=None, imgPath=None, xlabel="t/s"):
        """To draw the per frame statistic result on the sub graph. 
        If no subgraph is passed in, it will be displayed directly. 
        If imgpath is passed in, the graph will be saved. 

        start: 
            If xlabel is "t/s" or "t/ms", then it will be the starting time stamp. Default = 0. 
            If xlabel is "n", then it will be the starting sample count. Default = 0.
            If xlabel is "frame", then it will be the starting frame count. Default = 0.
        end: 
            If xlabel is "t/s" or "t/ms", then it will be the ending time stamp. Default = The last time stamp of the audio. 
            If xlabel is "n", then it will be the ending sample count. Default = The total count of samples. 
            If xlabel is "frame", then it will be the ending frame count. Default = The total count of frames. 
        ylim: A tuple (y_start,y_end) for display range on y-axis. The default (None) is adaptive. 
        ax: A matplotlib.pyplot subplot to draw on. 
        imgPath: The path to save the graph. 
        xlabel: "t/s", "t/ms", "n" or "frame". 
        """
        return super().plot(start, end, ylim, ax, imgPath, xlabel)
    
    def frameSelect(self, start = 0, end = None, step = None, index_type = "t/s"):
        """Selecting part of the result. 
        
        start: 
            If index_type is "t/s" or "t/ms", then it will be the starting time stamp of slicing. Default = 0. 
            If index_type is "n", then it will be the starting sample count of slicing. Default = 0.
            If index_type is "frame", then it will be the starting frame count. Default = 0.
        end: 
            If index_type is "t/s" or "t/ms", then it will be the ending time stamp of slicing. Default = The last time stamp of the audio. 
            If index_type is "n", then it will be the ending sample count of slicing. Default = The total count of samples. 
            If index_type is "frame", then it will be the ending frame count of slicing. Default = The total count of frames. 
        step: 
            If index_type is "t/s" or "t/ms", then it will be the time step of slicing. 
            If index_type is "n", then it will be the samples step of slicing.
            If index_type is "frame", then it will be the frames count step of slicing. 
            Default = 1 frame. 

        index_type: "t/s", "t/ms", "n" or "frame". 

        return:
            An 1-D NumPy array for part of the result. 
        """
        return self.samples[self.axis.slicing(start, end, step, index_type)]

class time_delta_axis_ds(one_axis_ds):
    def __init__(self, samples, sr, unit):
        super().__init__(samples, sr, time_delta_axis(len(samples), sr), unit)
    
    def plot(self, start=0, end=None, ylim=None, ax=None, imgPath=None, xlabel="t/s"):
        """To draw the auto correlation result on the sub graph. 
        If no subgraph is passed in, it will be displayed directly. 
        If imgpath is passed in, the graph will be saved. 

        start: 
            If xlabel is "t/s" or "t/ms", then it will be the starting time offset. Default = 0. 
            If xlabel is "k", then it will be the starting sample offset. Default = 0.
        end: 
            If xlabel is "t/s" or "t/ms", then it will be the ending time offset. Default = The duration of the audio. 
            If xlabel is "k", then it will be the ending sample offset. Default = The max samples count of the audio. 
        ylim: A tuple (y_start,y_end) for display range on y-axis. The default (None) is adaptive. 
        ax: A matplotlib.pyplot subplot to draw on. 
        imgPath: The path to save the graph. 
        xlabel: "t/s", "t/ms" or "k". 
        """
        return super().plot(start, end, ylim, ax, imgPath, xlabel)
    
    def timeSelect(self, start = 0, end = None, step = None, index_type = "t/s"):
        """Selecting part of the result. 
        
        start: 
            If index_type is "t/s" or "t/ms", then it will be the starting time offset of slicing. Default = 0. 
            If index_type is "k", then it will be the starting sample offset of slicing. Default = 0.
        end: 
            If index_type is "t/s" or "t/ms", then it will be the ending time offset of slicing. Default = The duration of the audio. 
            If index_type is "k", then it will be the ending sample offset of slicing. Default = The max samples count of the audio. 
        step: 
            If index_type is "t/s" or "t/ms", then it will be the time offset step of slicing. 
            If index_type is "k", then it will be the sample offset step of slicing.
            default = 1 sample offset. 

        index_type: "t/s", "t/ms" or "k". 

        return:
            An Audio object of part of the audio. 
        """
        return self.samples[self.axis.slicing(start, end, step, index_type)]

class freq_axis_ds_base(one_axis_ds):
    def __init__(self, samples, sr, unit):
        super().__init__(samples, sr, freq_axis(len(samples), sr), unit)
        self.scale_transform = {"linear": lambda a,b,c,d,e:e.slicing(a,b,c,d), "log": lambda a,b,c,d,e:e.logSlicing(a,b,c,d), "mel": lambda a,b,c,d,e:e.melSlicing(a,b,c,d)}
    
    def freqSelect(self, start = 0, end = None, step = None, index_type = "frequency/Hz"):
        """Selecting part of the spectrum. 
        
        start: The starting of slicing frequency. Meaning depends on index_type. Default = 0.
        end: The ending of slicing frequency. Meaning depends on index_type. Default = Half of the sample rate. 
        step: The slicing step. Meaning depends on index_type. Default = 1 freq point. 

        index_type: "frequency/Hz", "frequency/(rad/s)", "normalized frequency/Hz", "normalized frequency/(rad/s)" or "freq point". 

        return:
            An 1-D NumPy array for part of the spectrum. 
        """
        return self.samples[self.axis.slicing(start, end, step, index_type)]

class freq_axis_ds(freq_axis_ds_base):
    
    def plot(self, start=0, end=None, ylim=None, ax=None, imgPath=None, xlabel="frequency/Hz", plot_type="amplitude",freq_scale="linear"):
        """To draw the spectrogram on the sub graph. 
        If no subgraph is passed in, it will be displayed directly. 
        If imgpath is passed in, the graph will be saved. 

        start: The starting frequency. Meaning depends on xlabel. Default = 0. 
        end: The ending frequency. Meaning depends on xlabel. Default = Half of the sample rate. 
        ylim: A tuple (y_start,y_end) for display range on y-axis. The default (None) is adaptive. 
        ax: A matplotlib.pyplot subplot to draw on. 
        imgPath: The path to save the graph. 
        xlabel: "frequency/Hz", "frequency/(rad/s)", "normalized frequency/Hz", "normalized frequency/(rad/s)" or "freq point". 
        plot_type: "amplitude", "dB" or "phase". 
        freq_scale: "linear", "log" or "mel". 
        """
        plot_transform = {"amplitude": lambda x: np.abs(x), "dB": lambda x: librosa.amplitude_to_db(np.abs(x)), "phase": lambda x: np.angle(x)}
        
        return super().plot(start, end, ylim, ax, imgPath, xlabel, plot_transform[plot_type], slicing=self.scale_transform[freq_scale],unit=plot_type)

class freq_axis_ds_real(freq_axis_ds_base):
    def plot(self, start=0, end=None, ylim=None, ax=None, imgPath=None, xlabel="frequency/Hz",freq_scale="linear"):
        """To draw the statistic result spectrogram on the sub graph. 
        If no subgraph is passed in, it will be displayed directly. 
        If imgpath is passed in, the graph will be saved. 

        start: The starting frequency. Meaning depends on xlabel. Default = 0. 
        end: The ending frequency. Meaning depends on xlabel. Default = Half of the sample rate. 
        ylim: A tuple (y_start,y_end) for display range on y-axis. The default (None) is adaptive. 
        ax: A matplotlib.pyplot subplot to draw on. 
        imgPath: The path to save the graph. 
        xlabel: "frequency/Hz", "frequency/(rad/s)", "normalized frequency/Hz", "normalized frequency/(rad/s)" or "freq point". 
        freq_scale: "linear", "log" or "mel". 
        """
        return super().plot(start, end, ylim, ax, imgPath, xlabel, slicing=self.scale_transform[freq_scale])

class two_axis_ds:
    def __init__(self, samples, sr, axis1, axis2, unit):
        self.samples = samples
        self.sr = sr
        self.axis1 = axis1
        self.axis2 = axis2
        self.shape = samples.shape
        self.unit = unit
    
    def __getitem__(self, index):
        if type(index) == tuple:
            if len(index) > 2:
                raise Exception("Only support 1-D or 2-D indices. ")
            index1 = self.axis1[index[0]]
            index2 = self.axis2[index[1]]
            if type(index1) == np.array:
                index1 = index1[:,None]
            if type(index2) == np.array:
                index2 = index2[None,:]
            return self.samples[index1, index2]
        else:
            return self.samples[self.axis1[index]]
    
    def __len__(self):
        return self.samples.shape[0]
    
    def plot(self,xstart=0,xend=None,ystart=0,yend=None,
                    ax=None,imgPath=None,
                    xlabel="t/s",ylabel="t/s",
                    cbar=True,
                    plot_transform = lambda x:x,
                    slicing_axis1=lambda a,b,c,d,e:e.slicing(a,b,c,d),slicing_axis2=lambda a,b,c,d,e:e.slicing(a,b,c,d),unit=None):

        ax1_indice = slicing_axis1(xstart,xend,None,xlabel,self.axis1)
        ax1 = self.axis1.getScale(xlabel)[ax1_indice]
        ax2_indice = slicing_axis2(ystart,yend,None,ylabel,self.axis2)
        ax2 = self.axis2.getScale(ylabel)[ax2_indice]

        mat = self.samples[ax1_indice[:,None], ax2_indice[None,:]]
        mat = pd.DataFrame(plot_transform(mat.T),index=[round(i,2) for i in ax2],columns=[round(i,2) for i in ax1])
        if ax==None:
            plt.figure(figsize=(12, 4))
            sns.heatmap(mat,cmap=plt.cm.plasma,cbar=cbar)
            plt.xlabel(xlabel)
            plt.ylabel(ylabel)
            plt.xticks(rotation=60)
            # plt.gca().xaxis.set_major_formatter(StrMethodFormatter('{x:,.2f}'))
            # plt.gca().yaxis.set_major_formatter(StrMethodFormatter('{x:,.2f}'))
            if imgPath==None:
                plt.show()
            else:
                plt.savefig(imgPath,dpi=500, bbox_inches = 'tight')
            plt.clf()
        else:
            sns.heatmap(mat,cmap=plt.cm.plasma,xticklabels=ax1,yticklabels=ax2,ax=ax,cbar=cbar)

class time_frame_axis_ds(two_axis_ds):
    def __init__(self, samples, sr, step, unit):
        super().__init__(samples, sr, frame_axis(samples.shape[0], sr, step), time_axis(samples.shape[1], sr), unit)
    
    def frameSelect(self, start = 0, end = None, step = None, index_type = "t/s"):
        """Selecting part of the result on the frame axis. 
        
        start: 
            If index_type is "t/s" or "t/ms", then it will be the starting time stamp of slicing. Default = 0. 
            If index_type is "n", then it will be the starting sample count of slicing. Default = 0.
            If index_type is "frame", then it will be the starting frame count. Default = 0.
        end: 
            If index_type is "t/s" or "t/ms", then it will be the ending time stamp of slicing. Default = The last time stamp of the audio. 
            If index_type is "n", then it will be the ending sample count of slicing. Default = The total count of samples. 
            If index_type is "frame", then it will be the ending frame count of slicing. Default = The total count of frames. 
        step: 
            If index_type is "t/s" or "t/ms", then it will be the time step of slicing. 
            If index_type is "n", then it will be the samples step of slicing.
            If index_type is "frame", then it will be the frames count step of slicing. 
            Default = 1 frame. 

        index_type: "t/s", "t/ms", "n" or "frame". 

        return:
            An 2-D NumPy array for part of the result. 
        """
        return self.samples[self.axis1.slicing(start, end, step, index_type)]
    
    def timeSelect(self, start = 0, end = None, step = None, index_type = "t/s"):
        """Selecting part of the result on the time axis. 
        
        start: 
            If index_type is "t/s" or "t/ms", then it will be the starting time stamp of slicing. Default = 0. 
            If index_type is "n", then it will be the starting sample count of slicing. Default = 0.
        end: 
            If index_type is "t/s" or "t/ms", then it will be the ending time stamp of slicing. Default = The last time stamp in each frame. 
            If index_type is "n", then it will be the ending sample count of slicing. Default = The total count of samples in each frame. 
        step: 
            If index_type is "t/s" or "t/ms", then it will be the time step of slicing. 
            If index_type is "n", then it will be the samples step of slicing.
            Default = 1 sample. 

        index_type: "t/s", "t/ms" or "n". 

        return:
            An 2-D NumPy array for part of the result. 
        """
        return self.samples[:,self.axis2.slicing(start, end, step, index_type)]

class time_delta_frame_axis_ds(two_axis_ds):
    def __init__(self, samples, sr, step, unit):
        super().__init__(samples, sr, frame_axis(samples.shape[0], sr, step), time_delta_axis(samples.shape[1],sr), unit)
    
    def plot(self, xstart=0, xend=None, ystart=0, yend=None, ax=None, imgPath=None, xlabel="t/s", ylabel="t/s", cbar=True):
        """To draw the per frame statistic result on the sub graph. 
        If no subgraph is passed in, it will be displayed directly. 
        If imgpath is passed in, the graph will be saved. 

        xstart: 
            If xlabel is "t/s" or "t/ms", then it will be the starting time stamp. Default = 0. 
            If xlabel is "n", then it will be the starting sample count. Default = 0.
            If xlabel is "frame", then it will be the starting frame count. Default = 0.
        xend: 
            If xlabel is "t/s" or "t/ms", then it will be the ending time stamp. Default = The last time stamp of the audio. 
            If xlabel is "n", then it will be the ending sample count. Default = The total count of samples. 
            If xlabel is "frame", then it will be the ending frame count. Default = The total count of frames. 
        ystart: 
            If ylabel is "t/s" or "t/ms", then it will be the starting time offset. Default = 0. 
            If ylabel is "k", then it will be the starting sample offset. Default = 0.
        yend: 
            If ylabel is "t/s" or "t/ms", then it will be the ending time offset. Default = The duration of the audio in each frame. 
            If ylabel is "k", then it will be the ending sample offset. Default = The max samples count of the audio in each frame. 
        ax: A matplotlib.pyplot subplot to draw on. 
        imgPath: The path to save the graph. 
        xlabel: "t/s", "t/ms", "n" or "frame". 
        ylabel: "t/s", "t/ms" or "k"
        cbar: True to show the color bar. 
        """
        return super().plot(xstart, xend, ystart, yend, ax, imgPath, xlabel, ylabel,cbar)
    
    def frameSelect(self, start = 0, end = None, step = None, index_type = "t/s"):
        """Selecting part of the result on the frame dimension. 
        
        start: 
            If index_type is "t/s" or "t/ms", then it will be the starting time stamp of slicing. Default = 0. 
            If index_type is "n", then it will be the starting sample count of slicing. Default = 0.
            If index_type is "frame", then it will be the starting frame count. Default = 0.
        end: 
            If index_type is "t/s" or "t/ms", then it will be the ending time stamp of slicing. Default = The last time stamp of the audio. 
            If index_type is "n", then it will be the ending sample count of slicing. Default = The total count of samples. 
            If index_type is "frame", then it will be the ending frame count of slicing. Default = The total count of frames. 
        step: 
            If index_type is "t/s" or "t/ms", then it will be the time step of slicing. 
            If index_type is "n", then it will be the samples step of slicing.
             If index_type is "frame", then it will be the frames count step of slicing. 
            Default = 1 frame. 

        index_type: "t/s", "t/ms", "n" or "frame". 

        return:
            An 2-D NumPy array for part of the result. 
        """
        return self.samples[self.axis1.slicing(start, end, step, index_type)]
    
    def timeSelect(self, start = 0, end = None, step = None, index_type = "t/s"):
        """Selecting part of the result on the time offset dimension. 
        
        start: 
            If index_type is "t/s" or "t/ms", then it will be the starting time offset of slicing. Default = 0. 
            If index_type is "k", then it will be the starting sample differenc of slicing. Default = 0.
        end: 
            If index_type is "t/s" or "t/ms", then it will be the ending time offset of slicing. Default = The duration of the audio in each frame. 
            If index_type is "k", then it will be the ending sample differenc of slicing. Default = The max samples count of the audio in each frame. 
        step: 
            If index_type is "t/s" or "t/ms", then it will be the time offset step of slicing. 
            If index_type is "k", then it will be the sample offset step of slicing.
            default = 1 sample offset. 

        index_type: "t/s", "t/ms" or "k". 

        return:
            An 2-D NumPy array for part of the result. 
        """
        return self.samples[:,self.axis2.slicing(start, end, step, index_type)]

class freq_frame_axis_ds_base(two_axis_ds):
    def __init__(self, samples, sr, step, unit):
        super().__init__(samples, sr, frame_axis(samples.shape[0], sr, step), freq_axis(samples.shape[1], sr), unit)
        self.scale_transform = {"linear": lambda a,b,c,d,e:e.slicing(a,b,c,d), "log": lambda a,b,c,d,e:e.logSlicing(a,b,c,d), "mel": lambda a,b,c,d,e:e.melSlicing(a,b,c,d)}
    
    def frameSelect(self, start = 0, end = None, step = None, index_type = "t/s"):
        """Selecting part of the short-time spectrum on the frame dimension. 
        
        start: 
            If index_type is "t/s" or "t/ms", then it will be the starting time stamp of slicing. Default = 0. 
            If index_type is "n", then it will be the starting sample count of slicing. Default = 0.
            If index_type is "frame", then it will be the starting frame count. Default = 0.
        end: 
            If index_type is "t/s" or "t/ms", then it will be the ending time stamp of slicing. Default = The last time stamp of the audio. 
            If index_type is "n", then it will be the ending sample count of slicing. Default = The total count of samples. 
            If index_type is "frame", then it will be the ending frame count of slicing. Default = The total count of frames. 
        step: 
            If index_type is "t/s" or "t/ms", then it will be the time step of slicing. 
            If index_type is "n", then it will be the samples step of slicing.
             If index_type is "frame", then it will be the frames count step of slicing. 
            Default = 1 frame. 

        index_type: "t/s", "t/ms", "n" or "frame". 

        return:
            An 2-D NumPy array for part of the short-time spectrum. 
        """
        return self.samples[self.axis1.slicing(start, end, step, index_type)]
    
    def freqSelect(self, start = 0, end = None, step = None, index_type = "frequency/Hz"):
        """Selecting part of the short-time spectrum on the frequency dimension.  
        
        start: The starting of slicing frequency. Meaning depends on index_type. Default = 0.
        end: The ending of slicing frequency. Meaning depends on index_type. Default = Half of the sample rate. 
        step: The slicing step. Meaning depends on index_type. Default = 1 freq point. 

        index_type: "frequency/Hz", "frequency/(rad/s)", "normalized frequency/Hz", "normalized frequency/(rad/s)" or "freq point". 

        return:
            An 2-D NumPy array for part of the short-time spectrum. 
        """
        return self.samples[:,self.axis2.slicing(start, end, step, index_type)]

class freq_frame_axis_ds(freq_frame_axis_ds_base):
    
    def plot(self, xstart=0, xend=None, ystart=0, yend=None, ax=None, imgPath=None, xlabel="t/s", ylabel="frequency/Hz", plot_type="amplitude", cbar=True, freq_scale="linear"):
        """To draw the short-time spectrogram on the sub graph. 
        If no subgraph is passed in, it will be displayed directly. 
        If imgpath is passed in, the graph will be saved. 

        xstart: 
            If xlabel is "t/s" or "t/ms", then it will be the starting time stamp. Default = 0. 
            If xlabel is "n", then it will be the starting sample count. Default = 0.
            If xlabel is "frame", then it will be the starting frame count. Default = 0.
        xend: 
            If xlabel is "t/s" or "t/ms", then it will be the ending time stamp. Default = The last time stamp of the audio. 
            If xlabel is "n", then it will be the ending sample count. Default = The total count of samples. 
            If xlabel is "frame", then it will be the ending frame count. Default = The total count of frames. 
        ystart: The starting frequency. Meaning depends on ylabel. Default = 0. 
        yend: The ending frequency. Meaning depends on ylabel. Default = Half of the sample rate. 
        ax: A matplotlib.pyplot subplot to draw on. 
        imgPath: The path to save the graph. 
        xlabel: "t/s", "t/ms", "n" or "frame". 
        ylabel: "frequency/Hz", "frequency/(rad/s)", "normalized frequency/Hz", "normalized frequency/(rad/s)" or "freq point". 
        plot_type: "amplitude", "dB" or "phase". 
        cbar: True to show the color bar. 
        freq_scale: "linear", "log" or "mel". 
        """
        plot_transform = {"amplitude": lambda x: np.abs(x), "dB": lambda x: librosa.amplitude_to_db(np.abs(x)), "phase": lambda x: np.angle(x)}
        return super().plot(xstart, xend, ystart, yend, ax, imgPath, xlabel, ylabel, cbar, plot_transform[plot_type], slicing_axis2=self.scale_transform[freq_scale], unit=plot_type)
    
    

class freq_frame_axis_ds_real(freq_frame_axis_ds_base):
    def plot(self, xstart=0, xend=None, ystart=0, yend=None, ax=None, imgPath=None, xlabel="t/s", ylabel="frequency/Hz", cbar=True, freq_scale="linear"):
        """To draw the short-time statistic result spectrogram on the sub graph. 
        If no subgraph is passed in, it will be displayed directly. 
        If imgpath is passed in, the graph will be saved. 

        xstart: 
            If xlabel is "t/s" or "t/ms", then it will be the starting time stamp. Default = 0. 
            If xlabel is "n", then it will be the starting sample count. Default = 0.
            If xlabel is "frame", then it will be the starting frame count. Default = 0.
        xend: 
            If xlabel is "t/s" or "t/ms", then it will be the ending time stamp. Default = The last time stamp of the audio. 
            If xlabel is "n", then it will be the ending sample count. Default = The total count of samples. 
            If xlabel is "frame", then it will be the ending frame count. Default = The total count of frames. 
        ystart: The starting frequency. Meaning depends on ylabel. Default = 0. 
        yend: The ending frequency. Meaning depends on ylabel. Default = Half of the sample rate. 
        ax: A matplotlib.pyplot subplot to draw on. 
        imgPath: The path to save the graph. 
        xlabel: "t/s", "t/ms", "n" or "frame". 
        ylabel: "frequency/Hz", "frequency/(rad/s)", "normalized frequency/Hz", "normalized frequency/(rad/s)" or "freq point". 
        cbar: True to show the color bar. 
        freq_scale: "linear", "log" or "mel". 
        """
        return super().plot(xstart, xend, ystart, yend, ax, imgPath, xlabel, ylabel, cbar, slicing_axis2=self.scale_transform[freq_scale])


import numpy as np
from hmmlearn import hmm

def distance(x1, x2) :
    return np.sqrt(np.sum((x1-x2)**2))

def dtw(M1, M2) :
    '''Use DTW to calculate the similarity distance between two MFCC features. 

    M1: The first MFCC feature. 
    M2: The first MFCC feature. 

    return: A float object of the similarity distance between two MFCC features.
    '''
    M1_len = M1.shape[0]
    M2_len = M2.shape[0]
    cost = [[0 for i in range(M2_len)] for i in range(M1_len)]
    
    dis = []
    for i in range(M1_len) :
        dis_row = []
        for j in range(M2_len) :
            dis_row.append(distance(M1[i], M2[j]))
        dis.append(dis_row)
    
    cost[0][0] = dis[0][0]
    for i in range(1, M1_len) :
        cost[i][0] = cost[i - 1][0] + dis[i][0]
    for j in range(1, M2_len) :
        cost[0][j] = cost[0][j - 1] + dis[0][j]
    
    for i in range(1, M1_len) :
        for j in range(1, M2_len) :
            cost[i][j] = min(cost[i - 1][j] + dis[i][j] * 1, \
                            cost[i- 1][j - 1] + dis[i][j] * 2, \
                            cost[i][j - 1] + dis[i][j] * 1)
    return cost[M1_len - 1][M2_len - 1]

class GMMHMM:
    def __init__(self, features, labels, n_iter = 10):
        '''Construct and train a GMM+HMM model. 

        features: A list consisting of MFCC features.
        labels: The label corresponding to each MFCC feature in the features list. 
        n_iter: Iterating times. 

        return: A GMMHMM object. 
        '''
        GMMHMM_Models = {}
        states_num = 5
        GMM_mix_num = 6
        tmp_p = 1.0/(states_num-2)
        transmatPrior = np.array([[tmp_p, tmp_p, tmp_p, 0 ,0], [0, tmp_p, tmp_p, tmp_p , 0], [0, 0, tmp_p, tmp_p,tmp_p], [0, 0, 0, 0.5, 0.5], [0, 0, 0, 0, 1]],dtype=np.float)
        startprobPrior = np.array([0.5, 0.5, 0, 0, 0],dtype=np.float) 
        for label in np.unique(labels):
            trainData = [feature for feature, l in zip(features, labels) if l == label]
            model = hmm.GMMHMM(n_components=states_num, n_mix=GMM_mix_num, transmat_prior=transmatPrior, startprob_prior=startprobPrior, covariance_type='diag', n_iter=n_iter) 
            length = np.zeros([len(trainData), ], dtype=np.int) 
            for m in range(len(trainData)):
                length[m] = trainData[m].shape[0]
            trainData = np.vstack(trainData)
            model.fit(trainData, lengths=length)
            GMMHMM_Models[label] = model 
        self.GMMHMM_Models = GMMHMM_Models
    
    def predict(self, features):
        '''Use the trained GMM+HMM model to predict the labels on test set. 

        features: A list consisting of MFCC features.

        return: A list of predicted labels. 
        '''
        labels = []
        for feature in features:
            scoreList = {}
            for model_label in self.GMMHMM_Models.keys():
                model = self.GMMHMM_Models[model_label]
            score = model.score(feature)
            scoreList[model_label] = score
            predict = max(scoreList, key=scoreList.get)  
            labels.append(predict)
        return labels